#
# $XConsortium: Tekparse.def,v 1.2 88/09/06 14:35:12 jim Exp $
#
# List of symbols that need to be defined for Tekparse.h.  If you need to 
# change any of the CASE_ macros, make the change here and rerun the command
# shown in Tekparse.h.
#

CASE_REPORT
CASE_VT_MODE
CASE_SPT_STATE
CASE_GIN
CASE_BEL
CASE_BS
CASE_PT_STATE
CASE_PLT_STATE
CASE_TAB
CASE_IPL_STATE
CASE_ALP_STATE
CASE_UP
CASE_COPY
CASE_PAGE
CASE_BES_STATE
CASE_BYP_STATE
CASE_IGNORE
CASE_ASCII
CASE_APL
CASE_CHAR_SIZE
CASE_BEAM_VEC
CASE_CURSTATE
CASE_PENUP
CASE_PENDOWN
CASE_IPL_POINT
CASE_PLT_VEC
CASE_PT_POINT
CASE_SPT_POINT
CASE_CR
CASE_ESC_STATE
CASE_LF
CASE_SP
CASE_PRINT
CASE_OSC
