#
# $XConsortium: VTparse.def,v 1.4 92/09/15 15:28:31 gildea Exp $
# $Id: VTparse.def,v 6.1 1996/06/15 09:57:20 kagotani Rel $
#
# List of symbols that need to be defined for VTparse.h.  If you need to 
# change any of the CASE_ macros, make the change here and rerun the command
# shown in VTparse.h.
#

CASE_GROUND_STATE
CASE_IGNORE_STATE
CASE_IGNORE_ESC
CASE_IGNORE
CASE_BELL
CASE_BS
CASE_CR
CASE_ESC
CASE_VMOT
CASE_TAB
CASE_SI
CASE_SO
CASE_SCR_STATE
CASE_SCS0_STATE
CASE_SCS1_STATE
CASE_SCS2_STATE
CASE_SCS3_STATE
CASE_ESC_IGNORE
CASE_ESC_DIGIT
CASE_ESC_SEMI
CASE_DEC_STATE
CASE_ICH
CASE_CUU
CASE_CUD
CASE_CUF
CASE_CUB
CASE_CUP
CASE_ED
CASE_EL
CASE_IL
CASE_DL
CASE_DCH
CASE_DA1
CASE_TRACK_MOUSE
CASE_TBC
CASE_SET
CASE_RST
CASE_SGR
CASE_CPR
CASE_DECSTBM
CASE_DECREQTPARM
CASE_DECSET
CASE_DECRST
CASE_DECALN
CASE_GSETS
CASE_DECSC
CASE_DECRC
CASE_DECKPAM
CASE_DECKPNM
CASE_IND
CASE_NEL
CASE_HTS
CASE_RI
CASE_SS2
CASE_SS3
CASE_CSI_STATE
CASE_OSC
CASE_RIS
CASE_LS2
CASE_LS3
CASE_LS3R
CASE_LS2R
CASE_LS1R
CASE_PRINT
CASE_XTERM_SAVE
CASE_XTERM_RESTORE
CASE_XTERM_TITLE
CASE_DECID
CASE_HP_MEM_LOCK
CASE_HP_MEM_UNLOCK
CASE_HP_BUGGY_LL
CASE_TO_STATUS
CASE_FROM_STATUS
CASE_SHOW_STATUS
CASE_HIDE_STATUS
CASE_ERASE_STATUS
CASE_MBCS
CASE_SCS_STATE
